/*

Absolute format specifies an exact date and time when standard time
begins.  In this form, the wYear, wMonth, wDay, wHour, wMinute,
wSecond, and wMilliseconds members of the SYSTEMTIME structure are
used to specify an exact date.

Day-in-month format is specified by setting the wYear member to zero,
setting the wDayOfWeek member to an appropriate weekday, and using a
wDay value in the range 1 through 5 to select the correct day in the
month.  Using this notation, the first Sunday in April can be
specified, as can the last Thursday in October (5 is equal to the
last).

*/

#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include "listen.h"

#define SEC2MIN 60L
#define SEC2HOUR (60L*SEC2MIN)
#define SEC2DAY (24L*SEC2HOUR)

/* days per month plus days/year */
static	int	dmsize[]  = { 365, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
static	int	ldmsize[] = { 366, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

BOOL WINAPI GetTimeZoneInformationForYear(USHORT wYear, void *, TIME_ZONE_INFORMATION *);

/*
 * Windows doesn't seem to know the name of my daylight saving time, so we
 * have a fixed lookup table here. Maybe a similar problem for others?
 */
static struct {
	wchar_t *win;
	char *nine;
} Tznames[] = {
	{ L"GMT Standard Time",	"GMT" },
	{ L"GMT Daylight Time",	"BST" },
};

/* grotty heuristics to extract "GMT" from "GMT Standard Time"
 * 
 * http://social.msdn.microsoft.com/Forums/eu/vclanguage/thread/a87e1d25-fb71-4fe0-ae9c-a9578c9753eb
 * http://stackoverflow.com/questions/4195948/windows-time-zone-abbreviations-in-asp-net
 */
static void
guessname(char acronym[4], wchar_t *name)
{
	int i;
	char c[8];
	char *t;
	wchar_t *s;

	for(i = 0; i < sizeof(Tznames)/sizeof(Tznames[0]); i++)
		if(wcscmp(Tznames[i].win, name) == 0){
			strcpy(acronym, Tznames[i].nine);
			return;
		}

	/* use the first three capital letters (sigh) */
	for(s = name, t = acronym; *s; s++){
		wctomb(c, *s);
		if(isupper(*c))
			*t++ = *c;
		if((t - acronym) >= 3){
			*t = 0;
			return;
		}
	}
	strcpy(acronym, "???");
}

/*
 * day of week from year, month, and day.
 * uses Zeller's formula
 * http://mathforum.org/library/drmath/view/62324.html
 */
static int
weekday(int y, int m, int d)
{
	int adj, mm, yy;
 
	adj = (14-m) / 12;
	mm = m+12 * adj-2;
	yy = y-adj;
	return (d+ (13*mm-1) / 5+yy+yy / 4-yy / 100+yy / 400) % 7;
}

/* days/month for the given year */
static int *
yrsize(int y)
{
	if((y%4) == 0 && ((y%100) != 0 || (y%400) == 0))
		return ldmsize;
	else
		return dmsize;
}

/*
 * year, month, day of the week, and week number to seconds since epoch.
 * week number 5 indicates the last week of the month.
 */
static int
weekinmonth(int year, SYSTEMTIME *st)
{
	int secs, week, i, *d2m, yday, mday, firstday;

	secs = 0;
	yday = 0;
	week = st->wDay;			/* this field is used to hold week of month */

	/* previous years */
	for(i = 1970; i < year; i++){
		d2m = yrsize(i);
		secs += d2m[0] * SEC2DAY;
	}

	d2m = yrsize(year);

	/* previous months */
	for(i = 1; i < st->wMonth; i++)
		yday += d2m[i];

	/* full and partial weeks since start of month */
	firstday = weekday(year, st->wMonth, 1);
	if(firstday < st->wDayOfWeek)
		week--;
	mday = week*7 + (st->wDayOfWeek-firstday);
	if(week == 5 && mday >= d2m[st->wMonth])		/* wrap any overflows back a week */
		mday -= 7;

	secs += (yday+mday)*SEC2DAY;

	/* hours, minutes, seconds */
	secs += st->wHour * SEC2HOUR;
	secs += st->wMinute * SEC2MIN;
	secs += st->wSecond;

	if(secs < 0)
		return 0;
	return secs;
}


/* year, month, day, h, m, s to seconds since epoch */
static long
absday(int year, SYSTEMTIME *st)
{
	int i, *d2m, yday, secs;

	secs = 0;
	/*
	 *  seconds per year
	 */
	for(i = 1970; i < year; i++){
		d2m = yrsize(i);
		secs += d2m[0] * SEC2DAY;
	}

	yday = 0;
	d2m = yrsize(year);
	for(i = 0; i < st->wMonth-1; i++)
		yday += d2m[i+1];
	yday += st->wDay-1;

	secs += yday * SEC2DAY;

	/*
	 * hours, minutes, seconds
	 */
	secs += st->wHour * SEC2HOUR;
	secs += st->wMinute * SEC2MIN;
	secs += st->wSecond;

	if(secs < 0)
		return 0;
	return secs;
}

static int
tzgen(void)
{
	int yer, n;
	char buf[2048], stdname[4], dstname[4];
	TIME_ZONE_INFORMATION tzi;
	long summer, winter, spring, autumn;

	if(GetTimeZoneInformation(&tzi) == 0)
		return -1;

	winter = tzi.Bias * -60L;
	if(tzi.StandardDate.wMonth != 0)
		winter += tzi.StandardBias * -60;
	guessname(stdname, tzi.StandardName);

	if(tzi.DaylightDate.wMonth != 0 && tzi.DaylightBias != 0){
		guessname(dstname, tzi.DaylightName);
		summer = winter + tzi.DaylightBias * -60;
	}
	else{
		summer = winter;
		strcpy(dstname, stdname);
	}
	snprintf(buf, sizeof(buf), "%s %d %s %d\n", stdname, winter, dstname, summer);

	for(yer = 1986; yer < 2038; yer++){
		if(GetTimeZoneInformationForYear(yer, NULL, &tzi) == 0)
			continue;

		if(tzi.StandardDate.wYear)
			autumn = absday(yer, &tzi.StandardDate);
		else
			autumn = weekinmonth(yer, &tzi.StandardDate);

		/* the autumn date should be in std time not dst, so we adjust */
		autumn += tzi.DaylightBias * 60;

		if(tzi.DaylightDate.wYear)
			spring = absday(yer, &tzi.DaylightDate);
		else
			spring = weekinmonth(yer, &tzi.DaylightDate);

		n = strlen(buf);
		snprintf(buf+n, sizeof(buf)-n, "%ld %ld\n", spring, autumn);
	}

	SetEnvironmentVariable("timezone", buf);
	return 0;
}

void
initenv(void)
{
	DWORD n;
	SYSTEM_INFO si;
	char *plan9, *cpu, buf[128];
	static char ctldir[64], logfile[64];

	GetSystemInfo(&si);
	switch(si.wProcessorArchitecture){
	case PROCESSOR_ARCHITECTURE_AMD64:
		SetEnvironmentVariable("cputype", "win64");
		SetEnvironmentVariable("objtype", "win64");
		break;
	case PROCESSOR_ARCHITECTURE_INTEL:
		cpu = "win32";
		break;
	}
	snprintf(buf, sizeof(buf), "%lu", si.dwNumberOfProcessors);
	SetEnvironmentVariable("NPROC", buf);

	SetEnvironmentVariable("cputype", cpu);
	SetEnvironmentVariable("objtype", cpu);

	n = sizeof(buf);
	GetComputerName(buf, &n);
	SetEnvironmentVariable("sysname", buf);

	tzgen();

	if(Logfile && Ctldir)
		return;

	plan9 = getenv("plan9");
	if(plan9 == NULL){
		syslog("$plan9 not defined in environment");
		exit(1);
	}

	if(Logfile == NULL){
		snprintf(logfile, sizeof(logfile), "%s/sys/log/listen", plan9);
		Logfile = logfile;
	}

	if(Ctldir == NULL){
		snprintf(ctldir, sizeof(ctldir), "%s/%s/bin/service", plan9, cpu);
		Ctldir = ctldir;
	}
}

#ifdef TEST
char *Logfile;
char *Ctldir;
int Debug;
char *argv0;

void
syslog(char *fmt, ...)
{
	fmt = fmt;
}

void
main()
{
	tzgen();
}
#endif
