/* shell.c - spawn a shell and impose a plan9 style line discipline */

#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <winsock2.h>
#include <windows.h>
#include "dosd.h"

enum {
	Eof = 0x04,
	Dle = 0x10,
	Intr = 0x7f,
	Sub = 0x1a,

	Stdin_tout = 1000,		/* ms to wait if stdin fifo is full */
	Interrupt_grace = 100,	/* ms between sending ctrl-c and ctrl-break on interrupt */
	Hangup_grace = 1000,	/* ms between sending ctrl-c, ctrl-break and terminate on hangup */
};

/* line disipline */
static int
discipline(char *buf, int n, int sz, HANDLE child, int pid, HANDLE ph)
{
	int i, interrupted;
	static int esc = 0;
	static int brutality = 0;

	interrupted = 0;
	for(i = 0; i < n;){
		if(!esc && buf[i] == Dle){
			memmove(buf+i, buf+i+1, n-i);
			esc = 1;
			n--;
			continue;
		}
		if(esc && buf[i] == Intr){
			GenerateConsoleCtrlEvent(CTRL_C_EVENT, pid);
			if(WaitForSingleObject(ph, Interrupt_grace) == WAIT_TIMEOUT)
				GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, pid);
			if(brutality >= 3)
				if(WaitForSingleObject(ph, Interrupt_grace) == WAIT_TIMEOUT)
					TerminateProcess(child, 1);
			memmove(buf+i, buf+i+1, n-i);
			brutality++;
			interrupted++;
			n--;
			esc = 0;
			continue;
		}
		if(esc && buf[i] == Eof)
			return 0;
		esc = 0;
		i++;
	}
	if(interrupted == 0)
		brutality = 0;
	return n;
}

static void
relay(HANDLE src, HANDLE dst, HANDLE child, int pid, HANDLE ph)
{
	int failed;
	char buf[1024];
	HANDLE rx, tx, ha[2];
	unsigned long n, m;
	OVERLAPPED ol;

	rx = CreateEvent(NULL, FALSE, FALSE, NULL);
	WSAEventSelect((SOCKET)src, rx, FD_READ|FD_CLOSE);
	ha[0] = rx;
	ha[1] = child;

	tx = CreateEvent(NULL, FALSE, FALSE, NULL);
	memset(&ol, 0, sizeof(ol));
	ol.hEvent = tx;

	failed = 0;
	while(! failed) {
		if(WaitForMultipleObjects(2, ha, FALSE, INFINITE) != WAIT_OBJECT_0)
			break;

		if((n = recv((SOCKET)src, buf, sizeof(buf), 0)) <= 0){
			syswarn("inp: EOF on input\n");
			break;
		}

		if((n = discipline(buf, n, sizeof(buf), child, pid, ph)) < 0)
			break;

		if(WriteFile(dst, buf, n, &m, &ol) == 0){
			switch(GetLastError()){
			case ERROR_SUCCESS:
				break;
			case ERROR_IO_PENDING:
				if(WaitForSingleObjectEx(tx, Stdin_tout, TRUE) != WAIT_OBJECT_0)
					CancelIo(dst);
				break;
			default:
				failed = 1;
				syswarn("inp: write failed on kbd fifo (0x%x)\n", GetLastError());
				CancelIo(dst);
			}
		}
		FlushFileBuffers(dst);
	}
	CloseHandle(rx);
	CloseHandle(tx);

	if(WaitForSingleObject(ph, Hangup_grace) == WAIT_TIMEOUT)
		GenerateConsoleCtrlEvent(CTRL_C_EVENT, pid);
	if(WaitForSingleObject(ph, Hangup_grace) == WAIT_TIMEOUT)
		GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, pid);
	if(WaitForSingleObject(ph, Hangup_grace) == WAIT_TIMEOUT)
		TerminateProcess(child, 1);
}

/*
 * using named pipes allows us to make them message oriented.
 *
 * message orientated pipes can still read a byte at a time but
 * if we write a zero length message we get a zero length read,
 * which is an EOF by any other name.
 */
void
shell(HANDLE inp, HANDLE out, HANDLE err, HANDLE token, void *env, char *cmd)
{
	STARTUPINFO si;
	HANDLE inp_rd, inp_wr;
	PROCESS_INFORMATION pi;
	char *comspec, *plan9, *cputype;
	char npipe[MAX_PATH], cli[8192];

	WSADATA wsadata;
	if(WSAStartup(MAKEWORD(2, 2), &wsadata) != 0){
		syswarn("WSAStartup failed\n");
		return;
	}

	plan9 = getenv("plan9");
	comspec = getenv("COMSPEC");
	cputype = getenv("cputype");

	if(plan9 && cputype){
		if(*cmd)
			snprintf(cli, sizeof(cli), "%s/%s/bin/rc.exe -c \"%s\"", plan9, cputype, cmd);
		else
			snprintf(cli, sizeof(cli), "%s/%s/bin/rc.exe -l", plan9, cputype);
	}
	else
	if(comspec){
		syswarn("plan9=%s cputype=%s, incomplete environment, starting '%s'\n",
			((plan9)? plan9: "<nil>"), ((cputype)? cputype: "<nil>"), comspec);
		if(*cmd)
			snprintf(cli, sizeof(cli), "%s /q /a /d /f:off /c %s", comspec, cmd);
		else
			snprintf(cli, sizeof(cli), "%s /q /a /d /f:off /k", comspec);
	}
	else{
		syswarn("cannot find $plan9 and $cputype, or $COMSPEC, giving up");
		return;
	}

	snprintf(npipe, sizeof(npipe), "\\\\.\\pipe\\dosd.%d", GetCurrentProcessId());

	if((inp_wr = CreateNamedPipe(npipe, PIPE_ACCESS_OUTBOUND,
	  PIPE_TYPE_MESSAGE|PIPE_READMODE_BYTE,
	  PIPE_UNLIMITED_INSTANCES, 1024, 1024, 0, NULL)) == INVALID_HANDLE_VALUE)
		sysfatal("CreateNamedPipe() failed ");

	if((inp_rd = CreateFile(npipe, GENERIC_READ, 
	  0, NULL, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE)
		sysfatal("CreateFile(pipe) failed ");
	SetHandleInformation(inp_rd, HANDLE_FLAG_INHERIT, 1);

	memset(&si, 0, sizeof(si));
	si.cb = sizeof(STARTUPINFO);
	si.wShowWindow = SW_HIDE;
	si.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
	si.hStdInput = inp_rd;
	si.hStdOutput = out;
	si.hStdError = err;
	si.lpDesktop = "";		// we want a new winstation per session

	logmsg("cli: '%s'\n", cli);

	memset(&pi, 0, sizeof(pi));
	if(CreateProcessAsUser(token, NULL, cli, NULL, NULL, TRUE,
	    CREATE_UNICODE_ENVIRONMENT|CREATE_NEW_PROCESS_GROUP|CREATE_DEFAULT_ERROR_MODE,
	    env, NULL, &si, &pi) == 0)
		sysfatal("'%s' cannot create process", cli);

	CloseHandle(inp_rd);

	relay(inp, inp_wr, pi.hProcess, pi.dwProcessId, pi.hProcess);

	CloseHandle(inp_wr);
	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);
}
