/*
 * Naïve BSD LPR demon, see RFC 1179.
 *
 * Nice idea but doesn't work for windows network printers as these
 * need an authenticated session onto the print server. Works fine
 * for printers physically attached to the PC though.
 *
 * Bugs include
 *		- doesn't send a NAK on failures but just drops the line.
 */
#define _SPAWN_EXTENSION
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>

typedef struct Job Job;
struct Job {
	char *user;
	char *source;
	char *host;
	char *file;
	char content;
	int len;
};

char *plan9;
int Debug;

void
fatal(char *fmt, ...)
{
	char *msg;
	va_list ap;

	va_start(ap, fmt);
	msg = vsmprint(fmt, ap);
	va_end(ap);

	syslog(0, "print", "%s", msg);
	exits("failed");
}
void
ack(Biobuf *bp, int n)
{
	Bputc(bp, n);
	Bflush(bp);
}

int
ctrlfile(Biobuf *bi, char *spec, Job *j)
{
	int n, len;
	char *a[4], cmd, *args;

	if(tokenize(spec, a, nelem(a)) != 2)
		fatal("bad ctrl file spec");
	len = atoi(a[0]);

	while(len > 0){
		if((cmd = Bgetc(bi)) == -1)
			fatal("unexpected EOF in ctrl file");
		len--;

		args = Brdline(bi, '\n');
		n = Blinelen(bi);

		args[n-1] = 0;
		len -= n;

		/* These are only advisory so if strdup fails we don't care */
		switch(cmd){
		case 'N':
			j->source = strdup(args);
			break;
		case 'P':
			j->user = strdup(args);
			break;
		case 'H':
			j->host = strdup(args);
			break;
		default:
			if(islower(cmd))
				j->content = cmd;
			break;
		}
	}

	/* make logfile a little prettier */
	if(j->source == nil)
		j->source = strdup("-");
	return 0;	
}

int
datafile(Biobuf *bi, char *spec, Job *j)
{
	int n, m, len;
	Biobuf *bp;
	char *a[4], *file, buf[8192];

	if(tokenize(spec, a, nelem(a)) != 2)
		fatal("bad data file spec");
	len = atoi(a[0]);

	file = smprint("%s/tmp/lpr.XXXXXX", plan9);
	if(file == nil)
		fatal("No memory, %r");
	mktemp(file);
	if((bp = Bopen(file, OWRITE)) == nil)
		fatal("%s cannot create file, %r");

	j->len = len;
	j->file = file;

	while(len > 0){
		n = sizeof(buf);
		if(n > len)
			n = len;
		if((m = Bread(bi, buf, n)) <= 0)
			fatal("unexpected EOF in data file, %r");
		if(Bwrite(bp, buf, m) != n)
			fatal("%s write failed, %r", file);
		len -= m;
	}
	Bterm(bp);
	return 0;
}

void
freejob(Job *j)
{
	remove(j->file);

	free(j->user);
	free(j->source);
	free(j->host);
	free(j->file);
	memset(j, 0, sizeof(Job));
}

int
runjob(Job *j, char *spooler)
{
	Waitmsg *w;

	if(j->content != 'o'){
		syslog(0, "print", "%s '%c' is not postscript", j->source, j->content);
/*
 * Plan9 does not seem to flag postscript data as lpr expects.
 *		return -1;
 */
	}

	if(_spawnl(nil, spooler, spooler, "-q", j->file, nil) < 0){
		syslog(0, "print", "fail: '%s -q %s', %r", spooler, j->file);
		return 1;
	}
	w = wait();
	if(w)
		syslog(0, "print", "fail: '%s -q %s', %s", spooler, j->file, w->msg);
	else
		syslog(0, "print", "%s@%s %s %d", j->user, j->host, j->source, j->len);
	free(w);
	return 0;
}

int
submit(Biobuf *bi, Biobuf *bo, char *spooler)
{
	Job job, *j = &job;
	char cmd, *args;

	memset(j, 0, sizeof(Job));
	while((cmd = Bgetc(bi)) != -1){
		args = Brdline(bi, '\n');
		args[Blinelen(bi)-1] = 0;

		switch(cmd){
		case 1:						/* abort submission */
			freejob(j);
			break;
		case 2:						/* control file */
			ack(bo, 0);
			ctrlfile(bi, args, j);
			if((cmd = Bgetc(bi)) != 0)
				fatal("%d bad ctrl file terminator", cmd);
			break;
		case 3:						/* data file */
			ack(bo, 0);
			datafile(bi, args, j);
			if((cmd = Bgetc(bi)) != 0)
				fatal("%d bad data file terminator", cmd);
			break;
		default:
			fatal("%d bad submission request code", cmd);
			break;
		}

		if(j->file && j->content){
			runjob(j, spooler);
			freejob(j);
		}
		ack(bo, 0);
	}
	return 0;
}

void
main(void)
{
	char *spooler, cmd, *args;
	static Biobuf inp, *bi = &inp, out, *bo = &out;

	if((plan9 = getenv("plan9")) == nil)
		fatal("$plan9 not set");
	if((spooler = getenv("spool")) == nil)
		spooler = "c:/gs/gsview/gsprint.exe";

	Binit(bi, 0, OREAD);
	Binit(bo, 1, OWRITE);

	while((cmd = Bgetc(bi)) != -1){
		args = Brdline(bi, '\n');
		args[Blinelen(bi)-1] = 0;

		switch(cmd){
			case 1:					/* restart queue */
			goto done;
		case 2:						/* submit a job */
			ack(bo, 0);
			submit(bi, bo, spooler);
			break;
		case 3:						/* queue info short */
			Bprint(bo, "queue empty\n");
			goto done;
		case 4:						/* queue info long */
			Bprint(bo, "The printer queue is empty\n");
			goto done;
		case 5:						/* remove job */
			goto done;
		default:
			fatal("%d bad LPR request code", cmd);
			break;
		}
		Bflush(bo);
	}

done:
	exits(nil);
}

