/* print platform info (cpu load, num cpus etc) */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <win32.h>
#include <tlhelp32.h>

enum {
	SystemBasicInformation = 0,
	SystemPerformanceInformation = 2,
	SystemTimeInformation = 3,

	/* shorthand */
	Wks = VER_NT_WORKSTATION,
	Svr = VER_NT_SERVER,
	Dc = VER_NT_DOMAIN_CONTROLLER
};



#define ll2f(x)	((double)((x).HighPart) * 4.294967296E9 + (double)((x).LowPart))

typedef struct {
	LARGE_INTEGER	liKeBootTime;
	LARGE_INTEGER	liKeSystemTime;
	LARGE_INTEGER	liExpTimeZoneBias;
	ULONG uCurrentTimeZoneID;
	DWORD dwReserved;
} SYSTEM_TIME_INFORMATION;

typedef struct{
	LARGE_INTEGER	liIdleTime;
	DWORD		dwSpare[76];
} SYSTEM_PERFORMANCE_INFORMATION;

struct {
	int major;
	int minor;
	int type;
	char *name;
} Nt[] = {
	{ 6, 2, Wks, "8" },
	{ 6, 2, Svr, "server 2012" },
	{ 6, 1, Wks, "7" },
	{ 6, 1, Svr, "server 2008 r2" },
	{ 6, 0, Wks, "Vista" },
	{ 6, 0, Svr, "server 2008" },
	{ 5, 2, Svr, "server 2003" },
	{ 5, 2, Wks, "XP 64bit" },
	{ 5, 1, Wks, "XP" },
	{ 5, 0, Wks, "2000" },
	{ 4, 0, Wks, "NT" },
	{ 0, 0, 0, NULL }
};

void *
emalloc(n)
{
	char *p;

	if ((p = malloc((unsigned)n)) == nil)
		sysfatal("No memory");
	return p;
}

void *
erealloc(void *p, int n)
{
	char *q;

	if ((q = realloc(p, n)) == nil)
		sysfatal("no memory");

	return q;
}


void
cpuid(uint func, uint idx, uint *ax, uint *bx, uint *cx, uint *dx)
{
	int a, b, c, d;

	 __asm__ __volatile__ (
		"cpuid":
			"=a" (a), "=b" (b), "=c" (c), "=d" (d) :
			"a" (func), "c" (idx)
	);

	*ax = a;
	*bx = b;
	*cx = c;
	*dx = d;
}


static double
cpughz(void)
{
	HKEY hk;
	ulong mhz, sz;

	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE,
		L"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",
		0, KEY_READ, &hk) != 0)
		return -1;
	sz = sizeof(mhz);
	if(RegQueryValueEx(hk, L"~MHz", NULL, NULL, (void *)&mhz, &sz) != 0)
		return -1;
	return (double)mhz / 1000.0;
}

static char *
processor(void)
{
	HKEY hk;
	ulong sz;
	char *p, *name;
	wchar_t buf[1024];

	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE,
		L"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",
		0, KEY_READ, &hk) != 0)
		return "?";
	sz = nelem(buf);
	if(RegQueryValueEx(hk, L"ProcessorNameString", NULL, NULL, (LPBYTE)buf, &sz) != 0)
		return "?";

	name = _ninestr(buf);
	for(p = name; *p; p++)
		if(isspace(p[0]) && isspace(p[1]))
			strcpy(p, p+1);
	return name;
}

static double
load(int ncpus)
{
	int e;
	unsigned long stl, spl;
	double load, r1, r2, i1, i2;
	SYSTEM_TIME_INFORMATION *sti;         
	SYSTEM_PERFORMANCE_INFORMATION	*spi;
	LONG (WINAPI *qsi) (UINT, PVOID, ULONG, PULONG);

	stl = 1024;
	sti = emalloc(stl);
	qsi = latebind("ntdll", "NtQuerySystemInformation");
	while((e = qsi(SystemTimeInformation, sti, stl, &stl)) == 0xC0000004)
		sti = erealloc(sti, stl);
	if(e != NO_ERROR){
		fprint(2, "qsti failed sz=%d e=%x\n", stl, e);
		return -1;
	}
	r1 = ll2f(sti->liKeSystemTime);
	
	spl = 1024;
	spi = emalloc(spl);
	while((e = qsi(SystemPerformanceInformation, spi, spl, &spl)) == 0xC0000004)
		spi = erealloc(spi, spl);
	if(e != NO_ERROR){
		fprint(2, "SystemPerformanceInformation failed sz=%ld e=%lx\n", spl, e);
		return -1;
	}
	i1 = ll2f(spi->liIdleTime);

	Sleep(500);

	stl = 1024;
	sti = emalloc(stl);
	qsi = latebind("ntdll", "NtQuerySystemInformation");
	while((e = qsi(SystemTimeInformation, sti, stl, &stl)) == 0xC0000004)
		sti = erealloc(sti, stl);
	if(e != NO_ERROR){
		fprint(2, "SystemTimeInformation failed sz=%d e=%x\n", stl, e);
		return -1;
	}
	r2 = ll2f(sti->liKeSystemTime);
	
	spl = 1024;
	spi = emalloc(spl);
	while((e = qsi(SystemPerformanceInformation, spi, spl, &spl)) == 0xC0000004)
		spi = erealloc(spi, spl);
	if(e != NO_ERROR){
		fprint(2, "SystemPerformanceInformation failed sz=%ld e=%lx\n", spl, e);
		return -1;
	}
	i2 = ll2f(spi->liIdleTime);

	load = 100.0 - ((i2-i1) / ((r2-r1) * ncpus) * 100.0);
	if(load <= 0)
		return 0;
	return load;
}

int 
numhype()
{
	uint logical, cores, a, b, c, d;

	cpuid(1, 0, &a, &b, &c, &d);
	if(((d>>28)&1) == 0)
		return 1;
	logical = (b>>16)&0xff;

	cpuid(1, 0, &a, &b, &c, &d);
	if(a < 4)		/* leaf 4 not available */
		return 1;

	cpuid(4, 0, &a, &b, &c, &d);
	cores = ((a&0xfc000000)>>26)+1;

	return logical / cores;
}

int 
numcore()
{
	unsigned int a, b, c, d;

	cpuid(1, 0, &a, &b, &c, &d);
	if(a < 4)		/* leaf 4 not available */
		return 1;

	cpuid(4, 0, &a, &b, &c, &d);
	return ((a&0xfc000000)>>26)+1;
}

static char *
winver(SYSTEM_INFO *si, OSVERSIONINFOEX *ver)
{
	int i;
	static char buf[128];

	switch(ver->dwPlatformId){
	case VER_PLATFORM_WIN32s:
		return " win31";
	case VER_PLATFORM_WIN32_WINDOWS:
		if (ver->dwMinorVersion == 0)
			return "win95";
		else
			return "win98";
	case VER_PLATFORM_WIN32_NT:
		for(i = 0; Nt[i].name; i++)
			if(Nt[i].minor == ver->dwMinorVersion && 
			  Nt[i].major == ver->dwMajorVersion &&
			  Nt[i].type == ver->wProductType)
				return Nt[i].name;
		snprint(buf, sizeof(buf), " <%d.%d.%d>",
			ver->dwMinorVersion,
			ver->dwMajorVersion,
			ver->wProductType);
		return buf;
#ifdef VER_PLATFORM_WIN32_CE
	case VER_PLATFORM_WIN32_CE:
		return "winCE";
#endif
	}
	snprint(buf, sizeof(buf), " <%d>", ver->dwPlatformId);
	return buf;
}

static char *
cpuarch(SYSTEM_INFO *si)
{
	static char buf[128];

#ifndef PROCESSOR_AMD_X8664
#define PROCESSOR_AMD_X8664 8664
#endif

	switch(si->dwProcessorType){
#ifdef PROCESSOR_ARM
	case PROCESSOR_ARM:
		return "arm");
#endif
	case PROCESSOR_INTEL_386:
		return "i386";
	case PROCESSOR_INTEL_486:
		return "i486";
	case PROCESSOR_INTEL_PENTIUM:
		return "i586";
	case PROCESSOR_INTEL_IA64:
		return "ia64";
	case PROCESSOR_AMD_X8664:
		return "amd64";
	default:
		snprint(buf, sizeof(buf), " <%d>", si->dwProcessorType);
		return buf;
	}
}

char *
hostowner(void)
{
	typedef struct _WTS_PROCESS_INFO {
		DWORD SessionId;
		DWORD ProcessId;
		LPTSTR pProcessName;
		PSID pUserSid;
	} WTS_PROCESS_INFO;
	
	int i;
	SID_NAME_USE type;
	WTS_PROCESS_INFO *wp;
	wchar_t usr[128], dom[128];
	char *usrdom;
	ulong num, usrsz, domsz;
	void (*wtsfm)(void *);
	BOOL (*wtsep)(HANDLE, DWORD, DWORD, WTS_PROCESS_INFO **, DWORD *);

	wtsep = latebind("wtsapi32.dll", "WTSEnumerateProcessesW");
	wtsfm = latebind("wtsapi32.dll", "WTSFreeMemory");

	if(wtsep(0, 0, 1, &wp, &num) == 0)
		return "???";

	for(i = 0; i < num; i++)
		if(runestrcmp(wp[i].pProcessName, L"explorer.exe") == 0)
			break;
	if(i >= num)
		return "<none>";

	usrsz = nelem(usr);
	domsz= nelem(dom);
	if(LookupAccountSid(NULL, wp[i].pUserSid, usr, &usrsz, dom, &domsz, &type) == 0)
		return "<unknown>";
	wtsfm(wp);

	usrdom = _nineuser(usr, dom, 0);
	return usrdom;

}


void
main(int argc, char *argv[])
{
	int h;
	SYSTEM_INFO si;
	MEMORYSTATUSEX ms;
	OSVERSIONINFOEX ver;
	double memtot, loadave;
	BOOL WINAPI (*gmse)(MEMORYSTATUSEX *);

	argv0 = argv[0];

	debugpriv();
	gmse = latebind("kernel32.dll", "GlobalMemoryStatusEx");

	ver.dwOSVersionInfoSize = sizeof(ver);
	GetVersionEx((OSVERSIONINFO *)&ver);
	GetSystemInfo(&si);
	ms.dwLength = sizeof(ms);
	gmse(&ms);

	loadave = load(si.dwNumberOfProcessors);
	memtot = (double)ms.ullTotalPhys / (1024.0 * 1024.0 * 1024.0) + 0.05;

	h = numhype();
	if(h == 1)
		print("cores=%d\n", si.dwNumberOfProcessors);
	else
		print("cores=%d x %d\n", si.dwNumberOfProcessors, h);
	print("clock=%.1fGHz\n", cpughz() + 0.05);
	if(loadave < 0)
		print("cpuload=?\n");
	else
		print("cpuload=%.1f%%\n", loadave);
	print("memory=%.1fGb\n", memtot);
	print("memload=%.1f%%\n", (double)ms.dwMemoryLoad);
	print("os='Windows %s'\n", winver(&si, &ver));
	print("hostowner=%s\n", hostowner());
	print("arch='%s'\n", cpuarch(&si));
	print("cpu='%s'\n", processor());

	exits(nil);
}

