#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include <mswin.h>

#include <setupapi.h>

#include <ntdef.h>
#include <ddk/ntapi.h>
#include <ntapi-missing.h>

typedef struct Mount Mount;
struct Mount {
	char *dev;
	char *path;
	Mount *next;
};

static Mount *Mounttab;


void
add(wchar_t *dev, wchar_t *path)
{
	Mount *m, **p;
	char *devstr, *pathstr;

	devstr = _ninestr(dev);
	pathstr = _ninepath(path);

	p = &Mounttab;
	for(m = Mounttab; m; m = *p){
		p = &m->next;
		if(strcmp(m->dev, devstr) == 0 && strcmp(m->path, pathstr) == 0){
			free(devstr);
			free(pathstr);
			return;
		}
	}

	if((m = mallocz(sizeof(Mount), 1)) == NULL)
		sysfatal("No memory, %r");
	m->dev = devstr;
	m->path = pathstr;
	*p = m;
}


	
wchar_t *
reslashrune(wchar_t *s)
{
	wchar_t *p;

	if(s[1] == L':' && isupperrune(s[0]))	/* drive letter to lower case */
		s[0] = tolowerrune(s[0]);

	for(p = s; *p; p++)
		if(*p == L'\\'){					/* convert \\ to / */
			if(p[1] == 0)					/* strip trailing slashes */
				*p = 0;
			else
				*p = L'/';
		}
	return s;
}

wchar_t *
lowerrune(wchar_t *s)
{
	wchar_t *p;

	for(p = s; *p; p++)
		*p = tolowerrune(*p);
	return s;
}

void
rempath(wchar_t *dos)
{
	DWORD sz;
	wchar_t remote[1024];

	sz = nelem(remote);
	if(WNetGetConnection(dos, remote, &sz) == NO_ERROR)
		add(dos, remote);
}

void
devname(wchar_t *dos)
{
	wchar_t driver[1024];

	if(QueryDosDevice(dos, driver, nelem(driver)) >= 2)
		add(dos, driver);
}

int
proc_devmap(int pid, HANDLE h)
{
	DWORD got;
	int i, err;
	wchar_t dos[3];
	PROCESS_DEVICEMAP_INFORMATION dmi;

	got = sizeof(dmi);
	if((err = NtQueryInformationProcess(h, ProcessDeviceMap, &dmi, got, &got)) != STATUS_SUCCESS){
		fprint(2, "%s: get proc dev map failed, %s\n", argv0, _ntdllerr(err));
		return -1;
	}

	for (i = 0; i < sizeof(dmi.Query.DriveType); i++){
		dos[0] = L'A'+i;
		dos[1] = L':';
		dos[2] = 0;

		switch(dmi.Query.DriveType[i]){
			case DRIVE_UNKNOWN:
			case DRIVE_NO_ROOT_DIR:
				/* DFS volumes appear as NO_ROOT but are network drives */
				rempath(dos);
				devname(dos);
				break;
			case DRIVE_REMOVABLE:
				devname(dos);
				break;
			case DRIVE_FIXED:
				devname(dos);
				break;
			case DRIVE_CDROM:
				devname(dos);
				break;
			case DRIVE_RAMDISK:
				devname(dos);
				break;
			case DRIVE_REMOTE:
				rempath(dos);
				break;
			default:
				rempath(dos);
				break;
		}
	}
	return 0;
}

int
getwd_proc(Biobuf *bp, int pid, HANDLE h)
{
	PEB peb;
	DWORD got, err;
	wchar_t *buf;
	PROCESS_BASIC_INFORMATION pbi;
	RTL_USER_PROCESS_PARAMETERS upp;

	if((err = NtQueryInformationProcess(h, ProcessBasicInformation, &pbi, sizeof(pbi), &got)) != STATUS_SUCCESS){
		fprint(2, "%s: get basic proc info failed - %s\n", argv0, _ntdllerr(err));
		return -1;
	}

	if(ReadProcessMemory(h, pbi.PebBaseAddress, &peb, sizeof(peb), &got) == 0 || got != sizeof(peb)){
		fprint(2, "%s: cannot read PEB, %s\n", argv0, _winerr());
		return -1;
	}

	if(ReadProcessMemory(h, peb.ProcessParameters, &upp, sizeof(upp), &got) == 0 || got != sizeof(upp)){
		fprint(2, "%s: cannot read process params, %s\n", argv0, _winerr());
		return -1;
	}

	if((buf = malloc((upp.CurrentDirectoryName.MaximumLength+1) * sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");

	if(ReadProcessMemory(h, upp.CurrentDirectoryName.Buffer, buf, upp.CurrentDirectoryName.MaximumLength, &got) == 0 || got < 4){
		fprint(2, "%s: cannot read current dir, %s\n", argv0, _winerr());
		free(buf);
		return -1;
	}

	buf[got] = 0;
	Bprint(bp, "cd %w\n", reslashrune(buf));
	free(buf);
	return 0;
}

static void
walkreg(void *root, wchar_t *path)
{
	HKEY key;
	DWORD got;
	wchar_t *name, *value;
	int i, err, namsz, valsz;

	if(RegOpenKeyEx(root, path, 0, KEY_READ, &key) != ERROR_SUCCESS)
		return;

	namsz = 128;
	if((name = malloc(sizeof(wchar_t) * namsz)) == NULL)
		sysfatal("No memory, %r");

	valsz = 128;
	if((value = malloc(sizeof(wchar_t) * valsz)) == NULL)
		sysfatal("No memory, %r");

	i = 0;
	while(1){
		do{
			got = namsz;
			err = RegEnumValue(key, i++, name, &got, NULL, NULL, NULL, NULL);
			if(err == ERROR_MORE_DATA){
				namsz += 128;
				if((name = realloc(name, sizeof(wchar_t) * namsz)) == NULL)
					sysfatal("No memory, %r");
			}
		}while(err == ERROR_MORE_DATA);
		if(err != ERROR_SUCCESS)
			break;

		do{
			got = valsz;
			err = RegQueryValueEx(key, name, NULL, NULL, (BYTE *)value, &got);
			if(err == ERROR_MORE_DATA){
				valsz += 128;
				if((value = realloc(value, sizeof(wchar_t) * valsz)) == NULL)
					sysfatal("No memory, %r");
			}
		}while(err == ERROR_MORE_DATA);
		if(err != ERROR_SUCCESS)
			break;

		add(value, name);

	}
}


int
dosdevices()
{
	int len, attr;
	wchar_t *dos, *buf;

	buf = nil;
	for(len = 1024; len < 0xffff; len += 1024){
		if((buf = realloc(buf, len * sizeof(wchar_t))) == NULL)
			return 1;

		if(QueryDosDeviceW(nil, buf, len) > 0)
			break;

		if(GetLastError() != ERROR_INSUFFICIENT_BUFFER)
			sysfatal("QueryDosDevice(NULL) failed, %s", _winerr());
	}

	for(dos = buf; *dos; dos += runestrlen(dos)+1){
		attr = GetFileAttributes(dos);
		if(attr == INVALID_FILE_ATTRIBUTES)
			continue;

		rempath(dos);
		devname(dos);
	}

	free(buf);
	return 0;
}



int
LUIDenab(HANDLE h)
{
	int e;
	ULONG enab;

	enab = 0;
	if((e = NtQueryInformationProcess(h, ProcessLUIDDeviceMapsEnabled, &enab, sizeof(enab), nil)) != STATUS_SUCCESS){
		fprint(2, "%s: get LUID map enable flag failed, %s\n", argv0, _ntdllerr(e));
		return -1;
	}
	return enab;
}

static void
usage(void)
{
	fprint(2, "usage: %s [pid]\n", argv0);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int pid;
	HANDLE h;
	Mount *m;
	Biobuf bout;

	ARGBEGIN{
	}ARGEND;

	quotefmtinstall();
	_widefmtinstall();

	Binit(&bout, 1, OWRITE);

	if(_takepriv(SE_DEBUG_NAME) == -1)
		fprint(2, "%s: cannot gain debug priviledge: %r - some detail unavailable\n", argv0);

	switch(argc){
	case 0:
		pid = GetCurrentProcessId();
		break;
	case 1:
		pid = atoi(argv[0]);
		break;
	default:
		usage();
		break;
	}

	if((h = OpenProcess(PROCESS_ALL_ACCESS, 0, pid)) == nil)
		if((h = OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ, 0, pid)) == nil)
			if((h = OpenProcess(PROCESS_VM_READ, 0, pid)) == nil)
				if((h = OpenProcess(PROCESS_QUERY_INFORMATION, 0, pid)) == nil)
					sysfatal("%d cannot open process - %s\n", pid, _winerr());

	switch(LUIDenab(h)){
	case 0:								/* only global device mapping */
		walkreg(HKEY_LOCAL_MACHINE, L"HARDWARE\\DEVICEMAP\\SERIALCOMM");
		walkreg(HKEY_LOCAL_MACHINE, L"HARDWARE\\DEVICEMAP\\PARALLEL");
		dosdevices();
		break;
	case 1:								/* per-process device maps enabled */
		walkreg(HKEY_LOCAL_MACHINE, L"HARDWARE\\DEVICEMAP\\SERIALCOMM");
		walkreg(HKEY_LOCAL_MACHINE, L"HARDWARE\\DEVICEMAP\\PARALLEL");
		dosdevices();					/* non-disk devices only */
		proc_devmap(pid, h);			/* pre process device mapping */
		break;
	case -1:							/* cannot find out */
		break;
	}

	for(m = Mounttab; m; m = m->next)
		Bprint(&bout, "mount %-24s	%s\n", m->path, m->dev);

	getwd_proc(&bout, pid, h);
	exits(nil);
}
