/*
 * ps - process status list
 *
 * all usernames forced to lower case
 * whitespace in usernames replaced with
 *    underscores to keep collum count correct
 * extra -w option to aid win32 debugging
 * cannot get current priority
 */
#include <u.h>
#include <libc.h>
#include <bio.h>
#include <mswin.h>

#include <psapi.h>
#include <tlhelp32.h>
#include <imagehlp.h>

#include <wtsapi32.h>
#include <wtsapi32-missing.h>

#include <ntdef.h>
#include <ddk/ntapi.h>
#include <ntapi-missing.h>

typedef struct Proc Proc;
typedef struct Thread Thread;

struct Thread {
	Thread *next;
	HANDLE hand;		// thread handle
	int tid;			// thread ID
	int bprio;			// base priority
	int cprio;			// current priority
	char *state;		// thread state
	long rtime;			// elapsed real time
	long stime;			// system time
	long utime;			// user time
	void *entry;		// entrypoint of thread
	uvlong created;		// creation time
	vlong user;			// user time
	vlong kern;			// kernel time
	vlong create;		// creation time
	ulong ctxsw;		// context switch count
	vlong wait;			// time spent waiting or in the ready queue
};

struct Proc {
	Proc *next;
	HANDLE hand;		// process handle
	int pid;			// process ID
	int ppid;			// parent process ID
	int bprio;			// process base priority
	char *path;			// path to excutable
	char *name;			// process name
	char *args;			// complete command line
	char *owner;		// process owner
	vlong elap;			// elapsed real time
	vlong kern;			// kernel time
	vlong user;			// user time
	uvlong create;		// creation time
	vlong size;			// memory used
	int sess;			// windows session ID
	int stateid;		// numeric state ID
	char *state;		// textural state name
	Thread *threads;	// child threads
	int numhands;		// handle count

};

int Pflag = 0;		// pid and ppid
int pflag = 0;		// full command path and args
int aflag = 0;		// add priority info
int wflag = 0;		// add windows session ID
int rflag = 0;		// add realtime elapsed
int tflag = 0;		// add per-thread info
long Localnow;		// current time in secs since 1/1/70
int Verbose = 0;	// chatter
int Namlen = 0;		// max length of a user name

Proc *Procs;
char *Pathext;

void *
emallocz(int n, int c)
{
	void *v;

	if((v = mallocz(n, c)) == NULL)
		sysfatal("No memory, %r");
	return v;
}

char *
estrdup(char *p)
{
	char *q;

	if(p == NULL)
		return NULL;
	if((q = strdup(p)) == NULL)
		sysfatal("No memory, %r");
	return q;
}

void
elaptime(Biobuf *bp, long sec)
{
	long t;
	char buf1[32];

	if(sec == 0 || sec > Localnow)		// no info
		t = sec;
	else
		t = Localnow - sec;

	if(t >= 86400)
		snprint(buf1, sizeof(buf1), "%lud,%02lud:%02lud:%02lud", 
			t/86400, (t/3600)%24, (t/60)%60, t%60);
	else
	if(t >= 3600)
		snprint(buf1, sizeof(buf1), "%lud:%02lud:%02lud", t/3600, (t/60)%60, t%60);
	else
		snprint(buf1, sizeof(buf1), "%lud:%02lud", t/60, t%60);
	Bprint(bp, " %12s", buf1);
}

vlong
gint(HANDLE h, char *addr, int len)
{
	int i;
	DWORD got;
	uvlong vl, rc;
	unsigned char c;

	rc = 0;
	for (i = 0; i < len; i++){
		ReadProcessMemory(h, addr+i, &c, 1, &got);
		if(got != 1)
			return -1LL;
		vl = c;
		rc |= vl << (8*i);
	}
	return rc;
}



char *
statestr(long state, long wait)
{
	char *whystr[] = {
		"Executive",
		"FreePage",
		"PageIn",
		"PoolAlloc",
		"DelayExec",
		"Suspended",
		"UserRequest",
		"Executive",
		"FreePage",
		"PageIn",
		"PoolAlloc",
		"DelayExec",
		"Suspended",
		"UserReq",
		"EventPair",
		"Queue",
		"LPCrecv",
		"LPCreply",
		"VirtMem",
		"PageOut",
		"Rendez",
		"Spare2",
		"Spare3",
		"Spare4",
		"Spare5",
		"CalloutStk",
		"Kernel",
		"Resource",
		"PushLock",
		"Mutex",
		"QuantumEnd",
		"DispatchInt",
		"Preempted",
		"YieldExec",
		"FastMutex",
		"GuardMutex",
		"Rundown",
	};


	char *statestr[] = {
		"Initialised",
		"Ready",
		"Run",
		"Standby",
		"Term",
		"Wait",
		"SwapIn",
		"Unknown",
	};

	if(state == 5){
		if(wait >= 0 && wait < sizeof(whystr)/sizeof(whystr[0]))
			return smprint("%s", whystr[wait]);
		return smprint("Wait-%-2ld", wait);
	}
	if(state >= 0 && state < sizeof(statestr)/sizeof(statestr[0]))
		return smprint("%s", statestr[state]);
	return smprint("%2ld:%-2ld", state, wait);
}

/*
 * This is an educated guess, give the process a state 
 * related to the most interesting thread in it.
 */
void
procstate(Proc *p, int state, int wait)
{
	if(state != 5 || p->stateid == 0){			// Not a wait, or is just initialised
		p->state = statestr(state, wait);
		p->stateid = state;
	}
}

void
trimpath(char *name)
{
	char *p, *e, *ext;

	if((ext = strrchr(name, '.')) == nil)
		return;

	for(p = Pathext; p && *p; p = e+1){
		if((e = strchr(p, ' ')) == nil)
			if((e = strchr(p, ';')) == nil)		/* just in case rc is not expanded Pathext properly */
				e = strchr(p, 0);				/* only one extension in list */

		 if(strncmp(p, ext, (p-e)-1) == 0){
			*ext = 0;
			return;
		}
	}
}


Proc *
baseproc(SYSTEM_PROCESS_INFORMATION *spi)
{
	Proc **prv, *x, *p;

	if(spi->ProcessId == 0)
		return nil;

	p = emallocz(sizeof(Proc), 1);
	p->pid = spi->ProcessId;
	p->ppid = spi->InheritedFromProcessId;
	p->bprio = spi->BasePriority;
	p->size = spi->VmCounters.WorkingSetSize / 1024LL;
	p->user = _ft2elap((FILETIME *)&spi->UserTime) / 1000;
 	p->kern = _ft2elap((FILETIME *)&spi->KernelTime)/ 1000;
	p->create = _ft2elap((FILETIME *)&spi->CreateTime)/ 1000;
	p->numhands = spi->HandleCount;

	p->hand = OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ, 0, p->pid);
	if(p->hand == nil){
		p->hand = OpenProcess(PROCESS_QUERY_INFORMATION, 0, p->pid);
	}
	if(p->hand == nil){
		if(Verbose > 0)
			fprint(2, "OpenProcess(PROCESS_QUERY_INFORMATION, pid=%d) failed %s\n", p->pid, _winerr());
	}

	if(spi->ProcessName.Length > 0 && spi->ProcessName.Buffer != nil){
		p->path = _ninepath(spi->ProcessName.Buffer);
		p->name = estrdup(p->path);
	}
	else{
		p->path = estrdup("-");
		p->name = estrdup("-");
	}
	trimpath(p->name);

	prv = &Procs;
	for(x = Procs; x; x = x->next)
		prv = &x->next;
	*prv = p;
	return p;
}

Thread *
basethread(Proc *p, SYSTEM_THREAD_INFORMATION *sti)
{
	Thread **prv, *x, *t;

	t = emallocz(sizeof(Thread), 1);
	t->tid = (long)sti->ClientId.UniqueThread;		/* yes, really, the kernel handle is th usermode threadID */
	t->bprio = sti->BasePriority;
	t->cprio = sti->Priority;
	t->created = sti->CreateTime.QuadPart;
	t->state = statestr(sti->State, sti->WaitReason);
	t->entry = sti->StartAddress;
	t->wait = sti->WaitTime;
	t->ctxsw = sti->ContextSwitchCount;
	t->hand = OpenThread(THREAD_QUERY_INFORMATION, 0, t->tid);

	/* donate our state to owning proc */

	prv = &p->threads;
	for(x = p->threads; x; x = x->next)
		prv = &x->next;
	*prv = t;
	return t;
}

void
addthreaddetail(void)
{
	Proc *p;
	Thread *t;
	FILETIME c, e, k, u;

	for(p = Procs; p; p = p->next){
		for(t = p->threads; t; t = t->next){
			if(t->hand == BadHand)
				continue;
			if(GetThreadTimes(t->hand, &c, &e, &k, &u) == 0){
				continue;
			}
			t->user = _ft2elap(&u) / 1000;
			t->kern = _ft2elap(&k) / 1000;
			t->create = _ft2elap(&c) / 1000;

			NtQueryInformationThread(t->hand, ThreadQuerySetWin32StartAddress,
				&t->entry, sizeof(t->entry), NULL);
		}
	}
}

void
addargs(void)
{
	PEB peb;
	Proc *p;
	DWORD got;
	wchar_t *buf;
	int len, err, bufsiz;
	PROCESS_BASIC_INFORMATION pbi;
	RTL_USER_PROCESS_PARAMETERS upp;

	bufsiz = 1024;
	if((buf = malloc(bufsiz * sizeof(wchar_t))) == nil)
		sysfatal("No memory, %r");

	for(p = Procs; p; p = p->next){
		if(p->hand == nil)
			continue;

		if((err = NtQueryInformationProcess(p->hand, ProcessBasicInformation, &pbi, sizeof(pbi), &got)) != STATUS_SUCCESS || got != sizeof(pbi)){
			if(Verbose >1)
				print("pid=%d cannot get process basic info %s\n", _ntdllerr(err));
			continue;
		}
	
		if(ReadProcessMemory(p->hand, pbi.PebBaseAddress, &peb, sizeof(peb), &got) == 0 ||  got != sizeof(peb)){
			if(Verbose >1)
				fprint(2, "pid=%d cannot read PEB - %s", p->pid, _winerr());
			continue;
		}
		p->sess = peb.SessionId;	// works even on win2k without TS installed
	
		if(ReadProcessMemory(p->hand, peb.ProcessParameters, &upp, sizeof(upp), &got) == 0 || got != sizeof(upp)){
			if(Verbose >1)
				fprint(2, "pid=%d cannot read process parameters - %s", p->pid, _winerr());
			continue;
		}
	
		len = upp.CommandLine.MaximumLength;
		if(len >= bufsiz){
			bufsiz = len+1;
			if((buf = realloc(buf, bufsiz * sizeof(wchar_t))) == nil)
				sysfatal("No memory, %r");
		}

		if(ReadProcessMemory(p->hand, upp.CommandLine.Buffer, buf, len, &got) == 0 || got < 4){
			if(Verbose >1)
				print("%s:%d fail %s\n", __FILE__, __LINE__, _winerr());
			continue;
		}

		buf[got] = 0;
		p->args = _ninestr(buf);
	}
	free(buf);
}


void
addpaths(void)
{
	Proc *p;
	HANDLE hs;
	MODULEENTRY32 me;

	for(p = Procs; p; p = p->next){
		if((hs = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, p->pid)) == BadHand){
			if(Verbose >1)
				fprint(2, "CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, %d) failed - %s\n", p->pid, _winerr());
			continue;
		}
		me.dwSize = sizeof(me);
		if(Module32First(hs, &me) != 0)
			p->path = _ninepath(me.szExePath);
		else{
			if(Verbose >1)
				fprint(2, "Module32First(TH32CS_SNAPMODULE, %d) failed - %s\n", p->pid, _winerr());
		}
		CloseHandle(hs);
	}
}


/*
 * Try and open the process to get owner. This often fails
 * due to a lack of priviledge. Should only end up being
 * used on Win2k and NT4.
 */
void
addowners_open(void)
{
	Proc *p;
	HANDLE t;
	DWORD junk;
	TOKEN_USER *up;
	char *buf;
	enum {
		Bufsz = 8192,
	};

	if((buf = malloc(Bufsz)) == nil)
		sysfatal("No memory, %r");

	for(p = Procs; p; p = p->next){

		if(p->owner)
			continue;
		if(p->pid == 4){			/* we can never open pid 4 */
			p->owner = "System";
			p->sess = -1;
			continue;
		}

		if(p->hand == nil)
			continue;

		if(OpenProcessToken(p->hand, TOKEN_QUERY, &t) == 0){
			switch(GetLastError()){
			case ERROR_ACCESS_DENIED:
				break;
			default:
				if(Verbose >1)
					fprint(2, "OpenProcessToken(h, TOKEN_QUERY, pid=%d) failed - %s\n", p->pid, _winerr());
				break;
			}
			continue;
		}

		up = (TOKEN_USER *)buf;
		if(GetTokenInformation(t, TokenUser, up, Bufsz, &junk) == 0){
			fprint(2, "%s: pid=%d cannot get token information\n", argv0, p->pid);
			CloseHandle(t);
			continue;
		}
		p->owner = _sid2name(up->User.Sid, 100);
		p->sess = -1;
		CloseHandle(t);
	}
	free(buf);
}

/*
 * Use Terminal services to get owner (if available).
 * This is standard on WinXP, Vista and Win2k3 but
 * an option on Win2k and NT4 - 95/98/me are out of luck.
 */
void
addowners_wts(void)	
{
	int i;
	Proc *p;
	DWORD num;
	WTS_PROCESS_INFO *wp;

	wp = nil;
	if(WTSEnumerateProcesses(WTS_CURRENT_SERVER_HANDLE, 0, 1, &wp, &num) == 0)
		return;

	for(i = 0; i < num; i++){
		for(p = Procs; p; p = p->next){
			if(p->owner)
				continue;
			if(p->pid == 4){
				p->owner = "System";
				p->sess = -1;
				continue;
			}
			if(p->pid == wp[i].ProcessId){
				p->sess = wp[i].SessionId;
				if(wp[i].pUserSid != nil)
					p->owner = _sid2name(wp[i].pUserSid, 100);
				else{
					if(Verbose > 1)
						fprint(2, "WTSEnumerateProcesses(): pid=%d no process owner SID\n", p->pid);
				}
			}
		}
	}
	WTSFreeMemory(wp);
}

void
showthreads(Biobuf *bp, Proc *p)
{
	Thread *t;

	for(t = p->threads; t; t = t->next){

		Bprint(bp, "%*s ",  -Namlen, "");
		Bprint(bp, "%7d ", t->tid);
		if(Pflag)
			Bprint(bp, "%5s ", "");
		if(rflag)
			elaptime(bp, t->create);

		Bprint(bp, "%6lud:%.2lud %5lud:%.2lud ",
			t->user/60, t->user%60,
			t->kern/60, t->kern%60);
		if(pflag)
			Bprint(bp, "%2d %2d ", t->bprio, t->cprio);
		Bprint(bp, "%7s  ", "");
		if(wflag)
			Bprint(bp, "%2s ", "");
		Bprint(bp, "%-12s ", t->state);
		Bprint(bp, "%p\n", t->entry);
		
	}
}

void
showprocs(Biobuf *bp)
{
	Proc *p;

	for(p = Procs; p; p = p->next){
		Bprint(bp, "%*s ",  -Namlen, (p->owner)? p->owner: "-");
		Bprint(bp, "%7d ", p->pid);
		if(Pflag)
			Bprint(bp, "%7d ", p->ppid);
		if(rflag)
			elaptime(bp, p->create);

		Bprint(bp, "%6lud:%.2lud %5lud:%.2lud ",
			p->user/60, p->user%60,
			p->kern/60, p->kern%60);
		if(pflag)
			Bprint(bp, "%2d %2s ", p->bprio, "");
		Bprint(bp, "%7ludK ", p->size);
		if(wflag)
			Bprint(bp, "%2d ", p->sess);
		Bprint(bp, "%-12s ", p->state);

		if(aflag){
			if(p->args)
				Bprint(bp, "%s\n", p->args);
			else
			if(p->path)
				Bprint(bp, "%s\n", p->path);
			else
				Bprint(bp, "%s\n", p->name);
		}
		else{
			Bprint(bp, "%s\n", p->name);
		}
		if(tflag)
			showthreads(bp, p);
	}
}

void
widths(void)
{
	int n;
	Proc *p;

	Namlen = 0;
	for(p = Procs; p; p = p->next){
		n = strlen((p->owner)? p->owner: "-");
		if(n > Namlen)
			Namlen = n;
	}
}

void
usage(void)
{
	fprint(2, "usage: %s [-rpf]\n", argv0);
	fprint(2, "  -a         show all of the command line\n");
	fprint(2, "  -p         add base and current priorities\n");
	fprint(2, "  -P         add parent's process id\n");
	fprint(2, "  -r         add elapsed real time\n");
	fprint(2, "  -t         add thread detail\n");
	fprint(2, "  -v         verbose about problems\n");
	fprint(2, "  -w         add Windows session id\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Proc *p;
	char *buf;
	Biobuf bo;
	FILETIME ftnow;
	unsigned long sz;
	SYSTEMTIME sysnow;
	int nent, off, i, rc;
	SYSTEM_PROCESS_INFORMATION *spi;

	Verbose = 0;
	argv0 = argv[0];
	ARGBEGIN{
	case 'P':
		Pflag = 1;
		break;
	case 'r':
		rflag = 1;
		break;
	case 'p':
		pflag = 1;
		break;
	case 'a':
		aflag = 1;
		break;
	case 't':
		tflag = 1;
		break;
	case 'w':
		wflag = 1;
		break;
	case 'v':
		Verbose++;
		break;
	default:
		usage();
		break;
	}ARGEND;

	if(_takepriv(SE_DEBUG_NAME) == -1)
		fprint(2, "%s: cannot gain debug priviledge: %r - some detail unavailable\n", argv0);

	GetSystemTime(&sysnow);
	SystemTimeToFileTime(&sysnow, &ftnow);
	Localnow = _ft2elap(&ftnow) / 1000;
	Pathext = getenv("pathext");

	sz = 8192;
	buf = NULL;
	do{
		sz *= 2;
        if((buf = realloc(buf, sz)) == NULL)
			sysfatal("No memory, %r");
		rc = NtQuerySystemInformation(SystemProcessesAndThreadsInformation, buf, sz, NULL);
	}while(rc == STATUS_INFO_LENGTH_MISMATCH);
 	if(rc != NO_ERROR)
		sysfatal("NtQuerySystemInformation(SystemProcessesAndThreadsInformation) failed %s", _ntdllerr(rc));

  	off = 0;
	nent = 0;
	for(; spi = (SYSTEM_PROCESS_INFORMATION *)(buf+off), spi->NextEntryDelta != 0; off += spi->NextEntryDelta){
 		if((p = baseproc(spi)) == nil)
			continue;

		for(i = 0; i < spi->ThreadCount; i++, nent++){
			procstate(p, spi->Threads[i].State, spi->Threads[i].WaitReason);
			if(tflag)
				basethread(p, &spi->Threads[i]);
		}
	}

	addpaths();
	addowners_wts();
	addowners_open();
	if(aflag)
		addargs();
	if(tflag)
		addthreaddetail();
	widths();
	
	Binit(&bo, 1, OWRITE);
	showprocs(&bo);
	Bterm(&bo);
	exits(nil);
}
