#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <mswin.h>
#include "9win.h"

static int
rename(char *old, int srcattr, char *name)
{
	char *p, *new;
	wchar_t *src, *dst;
	int err, len, dstattr;

	len = strlen(old);
	if((new = malloc(len+MAX_PATH)) == nil)
		return -1;
	strcpy(new, old);

	if((p = strrchr(new, '/')) == nil && (p = strrchr(new, '\\')) == nil)
		p = new;
	else
		p++;
	strecpy(p, new+len+MAX_PATH, name);

	if(strcmp(old, new) == 0){
		free(new);
		return 0;
	}

	src = _winpath(old);
	dst = _winpath(new);
	free(new);

	dstattr = GetFileAttributes(dst);
	if(dstattr != -1 && (dstattr & FILE_ATTRIBUTE_READONLY))
		SetFileAttributes(dst, dstattr & ~FILE_ATTRIBUTE_READONLY);

	if(MoveFileEx(src, dst, MOVEFILE_REPLACE_EXISTING) == 0){
		err = GetLastError();
		SetFileAttributes(dst, dstattr);
		SetLastError(err);
		free(src);
		free(dst);
		return -1;
	}
	free(src);

	if(SetFileAttributes(dst, srcattr) == 0){
		free(dst);
		return -1;
	}
	free(dst);

	return 0;
}

static int
setmode(wchar_t *wpath, int attr, int mode)
{
	if(mode & 0222)
		attr &= ~FILE_ATTRIBUTE_READONLY;
	else
		attr |= FILE_ATTRIBUTE_READONLY;

	if(mode & DMTMP)
		attr |= FILE_ATTRIBUTE_TEMPORARY;
	else
		attr &= ~FILE_ATTRIBUTE_TEMPORARY;

	if(SetFileAttributes(wpath, attr) == 0)
		return -1;

	return 0;
}

static FILETIME 
epoch2ft(vlong vl)		/* expects milliseconds since 1/1/70 */
{
	FILETIME ft;

	vl += 11644473600000LL;
	vl *= 10000LL;
	ft.dwLowDateTime = vl;
	vl >>= 32;
	ft.dwHighDateTime = vl;
	return ft;
}


static int
settimes(wchar_t *wpath, long atime, long mtime)
{
	HANDLE h;
	FILETIME *acp, ac, *wrp, wr;

	if((h = CreateFile(wpath, FILE_FLAG_BACKUP_SEMANTICS|FILE_WRITE_ATTRIBUTES,
		0, nil, OPEN_EXISTING, 0, nil)) == BadHand)
		return -1;

	if(atime != ~0L || mtime != ~0L){
		acp = nil;
		if(atime != ~0L){
			ac = epoch2ft((vlong)atime*1000LL);
			acp = &ac;
		}
		wrp = nil;
		if(mtime != ~0L){
			wr = epoch2ft((vlong)mtime*1000LL);
			wrp = &wr;
		}

		if(SetFileTime(h, nil, acp, wrp) == 0){
			CloseHandle(h);
			return -1;
		}
	}
	CloseHandle(h);
	return 0;

}

static int
setlength(wchar_t *wpath, vlong length)
{
	HANDLE h;

	if((h = CreateFile(wpath, GENERIC_WRITE, 0, nil, OPEN_EXISTING, 0, nil)) == BadHand)
		return -1;

	if(length != ~0LL){
		if(SetFilePointerEx(h, (LARGE_INTEGER)length, nil, FILE_BEGIN) == 0){
			CloseHandle(h);
			return -1;
		}
		if(SetEndOfFile(h) == 0){
			CloseHandle(h);
			return -1;
		}
	}
	CloseHandle(h);
	return 0;

}

PSID 
GetSid(char *name)
{
	PSID sid;
	SID_NAME_USE snu;
	wchar_t *tmp, *dom;
	DWORD sidsz, domsz;

	sidsz = 0;
	domsz = 0;
	sid = nil;
	dom = nil;
	tmp = _winstr(name);
	if(LookupAccountName(nil, tmp, sid, &sidsz, dom, &domsz, &snu) == 0 &&
	  GetLastError() != ERROR_INSUFFICIENT_BUFFER){
		free(tmp);
		return nil;
	}

	if((sid = malloc(sidsz)) == nil){
		free(tmp);
		return nil;
	}
	if((dom = malloc(domsz * sizeof(wchar_t))) == nil){
		free(sid);
		free(tmp);
		return nil;
	}
	if(LookupAccountName(nil, tmp, sid, &sidsz, dom, &domsz, &snu) == 0){
		free(tmp);
		free(sid);
		free(dom);
		return nil;
	}
	free(dom);
	free(tmp);
	return sid;
}

int
setids(wchar_t *wpath, char *user, char *group)
{
	PSID usrsid, grpsid;
	int rc, which;
	uchar buf[1024];
	SECURITY_DESCRIPTOR *sd;

	rc = -1;
	which = 0;
	usrsid = nil;
	grpsid = nil;
	memset(buf, 0, sizeof(buf));
	sd = (SECURITY_DESCRIPTOR *)&buf;

	if(InitializeSecurityDescriptor(sd, SECURITY_DESCRIPTOR_REVISION) == 0)
		goto out;

	if(user && *user){
		if((usrsid = GetSid(user)) == nil)
			goto out;
		if(! IsValidSid(usrsid))
			goto out;
		if(SetSecurityDescriptorOwner(sd, usrsid, FALSE) == 0)
			goto out;
		which |= OWNER_SECURITY_INFORMATION;
	}

	if(group && *group){
		if((grpsid = GetSid(group)) == nil)
			goto out;
		if(SetSecurityDescriptorGroup(sd, grpsid, FALSE) == 0)
			goto out;
		which |= GROUP_SECURITY_INFORMATION;
	}

	if(SetFileSecurity(wpath, which, sd) == 0)
		goto out;

	rc = 0;
out:
	free(usrsid);
	free(grpsid);
	return rc;
}

int
_dowstat(char *path, Dir *d)
{
	wchar_t *tmp;
	int attr;

	tmp = _winpath(path);
	if((attr = _checkcase(tmp)) == -1)
		goto fail;

	if((d->uid && *d->uid) || (d->gid && *d->gid))
		if(setids(tmp, d->uid, d->gid) == -1)
			goto fail;

	if(d->muid && *d->muid){
		werrstr("'%s' muid not supported", path);
		free(tmp);
		return -1;
	}

	if(d->length != ~0LL)
		if(setlength(tmp, d->length) == -1)
			goto fail;

	if(d->atime != ~0L || d->mtime != ~0L)
		if(settimes(tmp, d->atime, d->mtime) == -1)
			goto fail;

	if(d->mode != ~0L){
		if(d->mode & DMAPPEND){
			werrstr("'%s' DMAPPEND not supported", path);
			free(tmp);
			return -1;
		}
		if(d->mode & DMEXCL){
			werrstr("'%s' DMEXCL not supported", path);
			free(tmp);
			return -1;
		}
		if(setmode(tmp, attr, d->mode) == -1)
			goto fail;
	}

	if(d->name && *d->name)
		if(rename(path, attr, d->name) == -1)
			goto fail;

	free(tmp);
	return 0;

fail:
	werrstr("'%s' %s", path, _winerr());
	free(tmp);
	return -1;

}
