#include <u.h>
#include <libc.h>
#include "error.h"
#include <mswin.h>
#include "9win.h"

#include <ntdef.h>
#include <winsock2.h>

/*
 * This is much more subtle than it first seems.
 *
 * Obviously his func needs to be recursive to walk the whole path,
 * but also it can only fail if the incorrect case exists but
 * the correct one does not - i.e. just because we cannot
 * find the file does not mean the filename is invalid;
 * consider nul: or c: for starters...
 *
 * we also need to correct the case of the cwd when the app starts
 * as this is cached in the proc table.
 *
 * It needs to be included in dostat(), open(), and chdir().
 * I don't beleive it is needed elsewhere but beware, lots of
 * testing is needed to get this right.
 */
int
_checkcase(wchar_t *tmp)
{
	HANDLE h;
	wchar_t *base, *p;
	WIN32_FIND_DATA fd;

	if((h = FindFirstFile(tmp, &fd)) == BadHand)
		return -1;
	CloseHandle(h);

	base = tmp;
	for(p = tmp; *p; p++)
		if(p[0] == L'/' || p[0] == L'\\' && p[1] != 0)
			base = p+1;

	if(_wincmp(fd.cFileName, base) != 0){
		SetLastError(ERROR_FILE_NOT_FOUND);
		return -1;
	}
	return fd.dwFileAttributes;
}

char *
_reslash(char *str)
{
	char *p, *l;

	str = strdup(str);
	if(str == nil)
		sysfatal("_reslash: No memory, %r");

	l = nil;
	for(p = str; *p; p++){
		if(*p == '\\')
			*p = '/';
		if(*p == '/')
			l = p;
	}

	/*
	 * has a slash,
	 * not the first letter (root dir),
	 * is last char,
	 * then clear it
	 */
	if(l != nil && l != str && l[1] == 0)
		*l = 0;

	return str;
}

static void
grow(Fgrp *fg, int new)
{
	int old;
		
	old = fg->nent;
	fg->nent = new;
	fg->ent = realloc(fg->ent, new*sizeof(File *));
	if(fg->ent == nil)
		_panic("allocfile/grow: No memory - %r\n");
	memset(fg->ent+old, 0, (new-old)*sizeof(File *));
}

static int
handtype(HANDLE h)
{
	DWORD m;
	int val, len;
	
	if(GetConsoleMode(h, &m) != 0)
		return Tcons;

	switch(GetFileType(h)){
	case FILE_TYPE_CHAR:
		return Tdev;
	case FILE_TYPE_DISK:
		return Tfile;
	case FILE_TYPE_PIPE:
		return Tpipe;
	case FILE_TYPE_UNKNOWN:
		if(GetLastError() == ERROR_INVALID_HANDLE)
			return Tinvalid;
		break;
	default:
		break;
	}

	if(_skt_getsockopt((SOCKET)h, SOL_SOCKET, SO_TYPE, (char *)&val, &len) == 0){
		if(val == SOCK_STREAM)
			return Tudp;
		if(val == SOCK_DGRAM)
			return Ttcp;
		_panic("unonown socket type");
	}
 
	return Tunknown;
}

/* always called with fgrp lock held */
int
_allocfile(Fgrp *fg, HANDLE hand, int type, int omode, char *path, int wantfd)
{
	File *f;
	int fd;

	if(type == Tunknown)				/* make an educated guess */
		type = handtype(hand);
	if(type == Tunknown){
		_debug("_allocfile: fd=%d hand=%p path=%s, unknown handle type, assuming Tfile\n", wantfd, hand, path);
		type = Tfile;
	}

	if((path = _reslash(path)) == nil)
		_panic("_allocfile: reslash failed - %r\n");

	if(wantfd == -1){
		for(fd = 0; fd < fg->nent; fd++)	/* use first available */
			if(fg->ent[fd] == nil)
				break;
		if(fd >= fg->nent){
			fd = fg->nent+1;
			grow(fg, fg->nent+32);
		}
	}
	else
	if(wantfd >= 0 && wantfd < fg->nent){	/* reuse a specific entry */
		fd = wantfd;
		if(fg->ent[fd] != nil)
			_freefile(fg, fd);
	}
	else{									/* allocate a specific entry */
		grow(fg, wantfd+32);
		fd = wantfd;
	}

	f = mallocz(sizeof(File), 1);
	if(f == nil)
		_panic("_allocfile: No memory - %r\n");

	f->hand = hand;
	f->type = type;
	f->omode = omode;
	f->path = path;
	f->off = 0;
	f->slot = 0;

	fg->ent[fd] = f;
	return fd;
}

File *
_lookfile(Fgrp *fg, int fd)
{
	File *f;

	if(fd < 0 || fd > fg->nent || fg->ent[fd] == nil){
		werrstr(Ebadfd);
		return nil;
	}
	f = fg->ent[fd];
	return f;
}

File *
_lookname(Fgrp *fg, char *path)			/* only used by accept() */
{
	int i;

	for(i = 0; i < fg->nent; i++)
		if(fg->ent[i] && strcmp(fg->ent[i]->path, path) == 0)
			return fg->ent[i];
	return NULL;
}

/* always called with fgrp locked */
int
_freefile(Fgrp *fg, int fd)
{
	File *f;

	if(fd < 0 || fd > fg->nent || fg->ent[fd] == nil){
		werrstr(Ebadfd);
		return -1;
	}
	f = fg->ent[fd];
	fg->ent[fd] = nil;
	free(f->path);
	free(f);

	return 0;
}

Fgrp *
_dupfgrp(Fgrp *old)
{
	int fd;
	Fgrp *new;
	HANDLE newhand;

	new = mallocz(sizeof(Fgrp), 1);

	new->nent = old->nent;
	new->ent = mallocz(new->nent * sizeof(File *), 1);
	if(new->ent == nil)
		_panic("newegrp: No memory - %r\n");

	for(fd = 0; fd < old->nent; fd++){
		if(old->ent[fd] == nil)
			continue;

		new->ent[fd] = calloc(sizeof(File), 1);
		if(new->ent[fd] == nil)
			_panic("newegrp: No memory - %r\n");

		if((DuplicateHandle(GetCurrentProcess(), old->ent[fd]->hand, GetCurrentProcess(), 
			&newhand, 0, FALSE, DUPLICATE_SAME_ACCESS)) == 0)
				_panic("newfgrp: cannot duplicate Fgrp handle - %r\n");

		if(_allocfile(new, newhand, old->ent[fd]->type, old->ent[fd]->omode, old->ent[fd]->path, fd) == -1)
			_panic("newfgrp: create copy Fgrp entry - %r\n");

		/* NB directory slot number is zero forcing a reread from the start */
		new->ent[fd]->off = old->ent[fd]->off;
	}
	return new;
}

void
_pushfgrp(Proc *p, Fgrp *fg)
{
	if(fg == nil)
		fg = mallocz(sizeof(Fgrp), 1);
	incref(fg);
	p = _getproc();
	fg->next = p->fgrp;
	p->fgrp = fg;
}

void
_popfgrp(Proc *p)
{
	int fd;
	Fgrp *fg;

	fg = p->fgrp;
	if(decref(fg) != 0)
		return;

	p->fgrp = fg->next;

	for(fd = 0; fd < fg->nent; fd++)
		_freefile(fg, fd);
	free(fg->ent);
	free(fg);
}


