#include <u.h>
#include <libc.h>
#include <mswin.h>

#include "9win.h"
#include "error.h"

enum {
	NoteOff = 16,
	CodeKill = 16,		/* pretend code for kill note */
};

int
notify(void (*f)(void *uregs, char *note))
{
	Proc *p;

	p = _getproc();
	if(p->notethread)
		CloseHandle(p->notethread);

	p->notify = f;
	if((p->notethread = OpenThread(THREAD_ALL_ACCESS, FALSE, GetCurrentThreadId())) == INVALID_HANDLE_VALUE)
		_panic("notify: cannot open this thread %s\n", _winerr());
	return 0;
}

int
noted(int ret)
{
	Proc *p;

	p = _getproc();
	longjmp(p->noted, ret+NoteOff);
}

/*
 * Exceptions are Async Proceedure Calls made from the kernel,
 * wether we are in alertable state or not, thus up-> should
 * relate to our proc struct as expected.
 */
LONG CALLBACK
_catchexcept(EXCEPTION_POINTERS *ep)
{
	Proc *p;
	Note *n;
	int rc, code;
	EXCEPTION_RECORD *er;

_debug("exception\n");
	if((p = _getproc()) == nil)
		_panic("no proc\n");

	if(ep == nil)
		_panic("catchexcept: no exception pointers\n");
	if((er = ep->ExceptionRecord) == nil)
		_panic("catchexcept: no exception record\n");
	code = ep->ExceptionRecord->ExceptionCode;

	if((n = _code2note(code)) == nil)
		_panic("_catchexcept: unknown exception code=0x%x", code);

_debug("exception %s\n", n->name);
	if(p->notify){
		rc = setjmp(p->noted);
		switch(rc){
		case 0:
			werrstr(n->name);
			p->notify((void *)ep, n->name);
			break;
		case NDFLT+NoteOff:
			break;
		case NCONT+NoteOff:
			if(er->ExceptionFlags & EXCEPTION_NONCONTINUABLE)
				sysfatal("continuing a non-continuable exception"); 
			SetEvent(p->rgrp->wait);		/* interrupt system call */
			p->notified = 1;
			if(strncmp(n->name, "sys: fp ", 8) == 0)
				_fprst(ep->ContextRecord);
			return 0;
		default:
			_panic("noted: %d invalid argument\n", rc);
		}
	}

	if(strncmp(n->name, "sys:", 4) == 0)
		_fault(ep, n->name);
	ExitProcess(1);
	return 0;			/* shut up gcc */
}

static VOID CALLBACK
interrupt(ULONG_PTR code)
{
	RaiseException((int)code, 0, 0, nil);
}

/*
 * Console events are a seperate thread created in our proc created
 * by the kernel, thus we must not use Thread Local Storage here.
 *
 * We can however get the first proc (of main()) from a static global variable.
 *
 * This is used to queue an async proceedure call to this main thread in the hope that
 * it will enter the alertable state (do a sleep, read or write) sometime soon
 * at which point interrupt() above will be called and cause an exception.
 */
BOOL WINAPI
_catchcons(DWORD code)
{
	Proc *p;

	p = _getmainproc();
	if(p->notify == nil){
		return 0; 		/* run the default handler */
	}

	if(QueueUserAPC(interrupt, p->notethread, (ULONG_PTR)(code+ConsExceptBase)) == 0)
		_panic("_catchcons: QueueUserAPC() failed %s\n", _winerr());
	return 1;			/* ignore */
}


static int
postone(int code, int pid)
{
	Proc *p;
	Note *n;
	HANDLE h;

	if((n = _code2note(code)) == nil){
		werrstr(Ebadarg);
		return -1;
	}

	switch(n->type){
	case Nterm:
		if((h = OpenProcess(PROCESS_TERMINATE, FALSE, pid)) == NULL){
			werrstr("%s", _winerr());
			return -1;
		}
		if(TerminateProcess(h, 1) == 0){
			werrstr("%s", _winerr());
			CloseHandle(h);
			return -1;
		}
		CloseHandle(h);
		break;
	case Ncons:
		if(GenerateConsoleCtrlEvent(code - ConsExceptBase, pid) != 0){
			werrstr("%s", _winerr());
			return -1;
		}
		break;
	case Nexcept:
		p = _getproc();
		if(pid != p->pid)
			_panic("postone: interprocess exceptions not supported\n");
			RaiseException((int)code, 0, 0, nil);
		break;
	default:
		_panic("postone: %d bad note type\n", n->type);
		break;
	}
	return 0;
}


int
postnote(int group, int pid, char *name)
{
	Note *n;

	if((n = _name2note(name)) == nil){
		werrstr("%s - unknown note", name);
		return -1;
	}

	switch(group){
	case PNPROC:
		if(postone(n->code, pid) < 0)
			return -1;
		break;
	case PNGROUP:
		/* FIXME: walk over proc table looking for processes
		 * in this processes group an send them a signal
		 */
		werrstr("postnote: note groups not supported");
		return -1;
		break;
	default:
		werrstr(Ebadarg);
		return -1;
	}
	return 0;
}

void
notejmp(void *uregs, jmp_buf env, int val)
{
	longjmp(env, val);
}

void
_pushngrp(Proc *p, Ngrp *ng)
{
	if(ng == nil)
		ng = mallocz(sizeof(Ngrp), 1);
	incref(ng);
	ng->next = p->ngrp;
	p->ngrp = ng;
}

void
_popngrp(Proc *p)
{
	Ngrp *ng;

	ng = p->ngrp;
	if(decref(ng) != 0)
		return;
#warning note groups unfinished

	p->ngrp = ng->next;
	free(ng);
}


