#include <u.h>
#include <libc.h>
#include <mswin.h>
#include "9win.h"

typedef struct Rendtag Rendtag;

struct Rendtag
{
	void *tag;
	void *val;
	Proc *proc;
	Rendtag *hash;
	Rendtag *free;
};

void *
rendezvous(void *tag, void *val)
{
	int e;
	uint h;
	Proc *p;
	void *ret;
	Rgrp *rg;
	Rendtag *t, **l;

	rg = _getproc()->rgrp;
	lock(rg);

	h = (ulong)tag & (NHASH-1);
	for(l=&rg->hashlst[h]; t=*l; l=&(*l)->hash){
		if(t->tag == tag){
			ret = t->val;
			t->val = val;
			p = t->proc;
			*l = t->hash;
			t->proc = nil;
			if(SetEvent(p->rgrp->wait) == FALSE)
				_panic("rendezvous: release failed %s\n", _winerr());
			/* lock passes to process being wakened */
			return ret;
		}
	}

	if((t = rg->freelst) == nil){
		t = mallocz(sizeof(*t), 1);
		if(t == nil)
			_panic("rendezvous: No memory - %r\n");
	}else
		rg->freelst = rg->freelst->free;

	t->tag = tag;
	t->val = val;
	t->hash = *l;
	t->proc = _getproc();
	*l = t;
	unlock(rg);

	/*
	 * there's a potential race here: if we get woken before we nap,
	 * the wake may be a no-op.  thus nap only naps for small
	 * amounts of time and we poll t->tag.  since we're using
	 * rendezvous as our synchronization primitive, i'm hesitant
	 * to build still other primitives and build rendezvous with them.
	 * losing this race is not at all likely: the time between
	 * unlocking the lock and napping is only a handful of instructions.
	 *
	 * depending on the implementation of pm_wake, the race may
	 * not be there at all: on some systems, pm_wake waits for the process
	 * to nap before waking it.
	 */
	while(t->proc != nil)
		if((e = WaitForSingleObject(t->proc->rgrp->wait, INFINITE)) != WAIT_OBJECT_0)
			break;

	/* lock inherited from waker, tag is unlinked from hash list */
	ret = t->val;
	t->free = rg->freelst;
	rg->freelst = t;
	unlock(rg);

	if(e != WAIT_OBJECT_0)
		ret = (void *)~0;
	return ret;
}

void
_pushrgrp(Proc *p, Rgrp *rg)
{
	if(rg == nil){
		rg = mallocz(sizeof(Rgrp), 1);
		if((rg->wait = CreateEvent(nil, FALSE, FALSE, nil)) == BadHand)
			_panic("_pushrgrp: cannot create event\n");
	}
	incref(rg);
	rg->next = p->rgrp;
	p->rgrp = rg;
}

void
_poprgrp(Proc *p)
{
	int i;
	Rgrp *rg;
	Rendtag *r, *nxt;

	rg = p->rgrp;

	qlock(rg);
	if(decref(rg) != 0){
		qunlock(rg);
		return;
	}

	p->rgrp = rg->next;

	SetEvent(rg->wait);
	CloseHandle(rg->wait);

	for(i = 0; i < NHASH; i++)
		for(r = rg->hashlst[i]; r; r = nxt){
			nxt = r->hash;
			free(r);
		}
	for(r = rg->freelst; r; r = nxt){
		nxt = r->free;
		free(r);
	}
	qunlock(rg);		/* lock hygiene */
	free(rg);
}


