#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <mswin.h>
#include "error.h"
#include "9win.h"

#include <shlwapi.h>

static wchar_t *
nine2win(wchar_t *win, int n, char *str, int ispath)
{
	Rune r;
	char *s;
	wchar_t *w;

	w = win;
	s = str;
	while(*s && w-win < n){
		if(ispath && *s == '/'){
			*w++ = L'\\';
			s++;
			continue;
		}
		s += chartorune(&r, s);
		*w++ = r;
	}
	*w = 0;

	/* trim trailing slashes, beware root dir */
	if(ispath)
		for(w--; w > win && *w == L'\\'; w--)
			*w = 0;

	return w+1;
}

static char *
win2nine(char *str, int n, wchar_t *win, int ispath)
{
	Rune r;
	wchar_t *w;
	char *s;

	s = str;
	w = win;


	while(*w && (s - str) < n){					/* wchar_t to utf8 and map \ to / */
		if(ispath && *w == L'\\'){
			*s++ = '/';
			w++;
			continue;
		}
		r = *w++;
		s += runetochar(s, &r);
	}
	assert(*w == 0);							/* str will always be long enough */
	*s = 0;

	if(ispath){									/* trim trailing slashes, beware root dir */
		for(s--; s > str && *s == '/'; s--)
			*s = 0;

		if(s[1] == ':' && isupper(s[0]))		/* any drive letter to lower case */
			s[0] = tolower(s[0]);
	}

	return s+1;
}

char *
_ninestr(wchar_t *win)
{
	int l;
	char *str;

	l = _winlen(win);
	if((str = malloc(l+1)) == nil)
		sysfatal("_ninestr: No memory, %r");
	win2nine(str, l, win, 0);
	return str;
}

/*
 * convert from wchar_t to utf8,
 * if the string is all caps, force to camel case
 * otherwise keep the existing case.
 * .e.e. stop windows from shouting like dos used to.
 */
static char *
quieten(char *s, wchar_t *w)
{
	Rune r, *t;
	int first, allcaps;

	allcaps = 1;
	for(t = w; *t; t++)
		if(islowerrune(*t)){
			allcaps = 0;
			break;
		}
	first = 1;
	while(*w){
		r = *w++;
		if(allcaps && isupperrune(r)){
			if(!first)
				r = tolowerrune(r);
			first = 0;
		}

		if(r != L' ')
			s += runetochar(s, &r);
		else
			first = 1;
	}
	return s;
}

/*
 * - we compress whitespace out of names for readibility
 * - if domain name exists (and is interesting) we append
 *   it to the username seperated by an '@'
 * - We force names in all caps into CamelCase
 *   if domain name exists
 * - uninteresting domain names will be stripped.
 * - we always strip domain names that are the same as our
 *   login domain.
 * - In agressive mode we also strip well-known domain names
 *   and the name of our machine as a domain name.
 */
char *
_nineuser(wchar_t *dom, wchar_t *usr)
{
	int l;
	char *s, *str;

	l = _winlen(usr);
	if(dom && *dom)
		l += _winlen(dom) +1;

	if((str = malloc(l+1)) == nil)
		sysfatal("_nineuser: No memory, %r");

	s = quieten(str, usr);
	if(dom && *dom){
		*s++ = '@';
		s = quieten(s, dom);
	}
	*s = 0;
	return str;
}

char *
_ninepath(wchar_t *win)
{
	int l;
	char *str;

	l = _winlen(win);
	if((str = malloc(l+1)) == nil)
		sysfatal("_ninepath: No memory %r");
	win2nine(str, l, win, 1);
	return str;
}

char *
_nineargv(wchar_t *cli, int *acp, char ***avp)
{
	int l, ac;
	wchar_t *w;
	char *buf, **av;

	l = _winlen(cli) + 1;
	ac = 10;						/* plenty extra */
	for(w = cli; *w; w++)
		if(*w == L' ' || *w == L'\t')
			ac++;

	av = malloc(l + ac * sizeof(char *));
	buf = (char *)&av[ac];			/* just off the end */
	win2nine(buf, l, cli, 0);
	*acp = ac;
	*avp = av;
	return buf;
}

/***********************************************/

int
_wincmp(wchar_t *s, wchar_t *t)
{
	if (s == t)
		return 0;

	while(*s && *s == *t){
		s++;
		t++;
	}

	return (int)*s - (int)*t;
}

int
_winlen(wchar_t *w)
{
	int n;

	n = 0;
	while(*w)
		n += runelen(*w++);
	return n;
}


wchar_t *
_winstr(char *s)
{
	int l;
	wchar_t *win;

	l = utflen(s);
	if((win = malloc((l+1)*sizeof(wchar_t))) == nil)
		sysfatal("_winstr: No memory, %r");
	nine2win(win, l, s, 0);
	return win;
}

int
_windevice(char *name)
{
	int i;
	char a, b, *s, *t;
	static char *devs[] = { "nul", "prn", "con", "lpt", "com" };

	for(i = 0; i < nelem(devs); i++){
		s = name;
		t = devs[i];

		for(; *s; s++, t++){
			a = tolower(*s);
			b = tolower(*t);
			if(a != b)
				break;
		}
		if(*s == 0)
			return 1;						/* exact match */

		if(*t == 0){						/* name matches but trailing stuff */
			while(isdigit(*s))				/* strip device instance number */
				s++;
			if(*s == ':')					/* strip trailing colon */
				s++;
			if(*s == 0)
				return 1;
		}
	}
	return 0;
}

static int
abspath(char *s)
{
	if(*s == '/' || *s == '\\')			/* root dir */
		return 1;
	if(isalpha(s[0]) && s[1] == ':')	/* drive letter */
		return 1;
	return -1;
}

/*
 * Windows APIs are inconsistent (shock).
 * 	- some insist on a drive letter "x:" rather than just "/"
 *	- some insist on "x:" rather than "x:/"
 */
static void
bodge(char *s, char *cwd)
{
	if(strcmp(s, "/") == 0){
		s[0] = cwd[0];
		s[1] = ':';
		s[2] = 0;
	}

	if(isalpha(s[0]) && strcmp(s+1, ":/") == 0)
		s[2] = 0;
}

/* NB: the filename arg should not contain path seperators */ 
wchar_t *
_winpath2(char *path, wchar_t *filename)
{
	int l;
	Proc *p;
	char *tmp;
	wchar_t *w, *win;

	p = _getproc();
	l = utflen(path)+1;
	if(abspath(path)){
		if((tmp = strdup(path)) == NULL)
			sysfatal("_winpath: No memory, %r");
	}
	else{
		l = strlen(path)+1 + strlen(p->cwd)+1;
		if((tmp = malloc(l)) == nil)
			sysfatal("_winpath: No memory, %r");
		strcpy(tmp, p->cwd);
		strcat(tmp, "/");
		strcat(tmp, path);
	}

//	cleanname(tmp);
	bodge(tmp, p->cwd);

	l = utflen(tmp);
	if(filename)
		l += _winlen(filename) +1;

	if((win = malloc((l+1)*sizeof(wchar_t))) == nil)
		sysfatal("_winstr: No memory, %r");
	w = nine2win(win, l, tmp, 1);
	free(tmp);

	if(filename){
		*w++ = L'\\';
		memcpy(w, filename, sizeof(wchar_t) * (_winlen(filename)+1));
	}
	return win;
}

wchar_t *
_winpath(char *path)
{
	return _winpath2(path, nil);
}

wchar_t *
_winenv(char *str)
{
	int l, n;
	char *s;
	wchar_t *w, *win;

	for(s = str; *s; s++)
		s += strlen(s)+1;
	n = (s-str)+1;

	if((win = malloc(n*sizeof(wchar_t))) == nil)
		sysfatal("_winenv: No memory, %r");

	w = win;
	for(s = str; *s; s++){
		l = strlen(s);
		nine2win(w, l, s, 0);
		w += utflen(s);
		*w++ = 0;
		s += l;
	}
	*w = 0;
	return win;
}

