#include <u.h>
#include <libc.h>
#include "error.h"
#include <mswin.h>
#include "9win.h"

typedef struct {
	int e;
	char *s;
} Errmsg;

/* some more common errors translated to more friendly messages */
static Errmsg Winerr[] = {
	{ ERROR_ACCESS_DENIED,			Eperm },
	{ ERROR_ALREADY_EXISTS,			Eexist },
	{ ERROR_BAD_EXE_FORMAT,			Ebadexec },
	{ ERROR_BAD_NETPATH,			"network path not found" },	/* e.g. broken DFS path */
	{ ERROR_BAD_NET_NAME,			"host not known" },			/* e.g. broken DFS host */
	{ ERROR_BAD_PATHNAME,			Efilename },
	{ ERROR_BROKEN_PIPE,			"broken pipe" },
	{ ERROR_DIR_NOT_EMPTY,			"directory not empty" },
	{ ERROR_FILE_NOT_FOUND,			Enonexist },
	{ ERROR_GEN_FAILURE,			Eio },
	{ ERROR_HANDLE_DISK_FULL,		"no space on device" },
	{ ERROR_HANDLE_EOF,				"end of file" },
	{ ERROR_INVALID_HANDLE,			"invalid handle" },
	{ ERROR_INVALID_NAME,			Efilename },
	{ ERROR_NEGATIVE_SEEK,			Enegoff },
	{ ERROR_NONE_MAPPED,			"name not unknwon" },		/* LookupAccountName() on bad name */
	{ ERROR_NOT_ENOUGH_MEMORY,		Enovmem },
	{ ERROR_NOT_READY,				"device not ready" },		/* e.g. cdrom with no disk */
	{ ERROR_NOT_SUPPORTED,			"not supported" },
	{ ERROR_NO_MORE_SEARCH_HANDLES,	"too many open directories" },
	{ ERROR_OUTOFMEMORY,			Enomem },
	{ ERROR_PATH_NOT_FOUND,			Enonexist },
	{ ERROR_SEEK_ON_DEVICE,			Eisstream },
	{ ERROR_SHARING_VIOLATION,		"sharing violation" },
	{ ERROR_SUCCESS,				Enoerror },
	{ ERROR_TOO_MANY_OPEN_FILES,	Enofd },
	{ ERROR_WRITE_PROTECT,			"read only" },
	{ ERROR_INVALID_PARAMETER,		Ebadarg },
	{ ERROR_PARTIAL_COPY,			"VM read/write incomplete" },
};

void
_cleanerr(char *msg)
{
	int f;
	char *r, *w;

	f = 0;
	for(r = msg, w = msg; *r; r++)
		switch(*r){
		case '\r':
		case '\n':
		case ' ':
			if(f++ == 0)
				*w++ = ' ';
			break;
		default:
			f = 0;
			*w++ = *r;
			break;
		}
	*w = 0;
	while(w > msg && (*w == ' ' || *w == '.' || *w == 0))
		*w-- = 0;
}

static int
fmtmsg(char *buf, int len, int err, HMODULE lib, char *prefix)
{
	int flags;
	char *str;
	wchar_t msg[ERRMAX];

	if(len)
		*buf = 0;
	if(err == 0)
		return 0;

	flags = FORMAT_MESSAGE_MAX_WIDTH_MASK;
	if(lib)
		flags |= FORMAT_MESSAGE_FROM_HMODULE;
	else
		flags |= FORMAT_MESSAGE_FROM_SYSTEM;

	if(FormatMessage(flags, lib, err, 0, msg, nelem(msg), nil) == 0)
		return -1;

	str = _ninestr(msg);
	_cleanerr(str);
	snprint(buf, len, "%s %s (0x%ux)", prefix, str, err);
	free(str);
	return 0;
}

char *
_winerr(void)
{
	int i, e;
	static char buf[ERRMAX];

	*buf = 0;
	e = GetLastError();
	SetLastError(0);
	if(e == 0)
		return buf;

	for(i = 0; i < nelem(Winerr); i++)
		if(e == Winerr[i].e)
			return Winerr[i].s;

	if(fmtmsg(buf, sizeof(buf), e, nil, "") == -1)
		snprint(buf, sizeof(buf), "%d unknown error code", e);
	return buf;
}

char *
_ntdllerr(uint e)
{
	int i, rc;
	HMODULE lib;
	char *prefix;
	static char buf[ERRMAX];

	*buf = 0;
	switch((e >> 30) & 3){
	case 0:
		prefix = "ok: ";
		break;
	case 1:
		prefix = "info: ";
		break;
	case 2:
		prefix = "warn: ";
		break;
	case 3:
		prefix = "err: ";
		break;
	default:
		prefix = "";
		break;
	}

	switch(e){
	case 0xC0000022:	return "permission denied";
	case 0xc000000d:	return "bad parameter";
	}

	lib = LoadLibrary(L"ntdll.dll");
	rc = fmtmsg(buf, sizeof(buf), e, lib, prefix);
	FreeLibrary(lib);

	if(rc == -1){
		for(i = 0; i < nelem(Winerr); i++)
			if(e == Winerr[i].e)
				return Winerr[i].s;
		fmtmsg(buf, sizeof(buf), e, nil, "");
	}
	return buf;
}
	

